@@ -3,9 +3,9 @@
 # Univention Python
 #  LDAP access
 #
-# Copyright 2002-2019 Univention GmbH
+# Copyright 2002-2021 Univention GmbH
 #
-# http://www.univention.de/
+# https://www.univention.de/
 #
 # All rights reserved.
 #
@@ -28,11 +28,13 @@
 # You should have received a copy of the GNU Affero General Public
 # License with the Debian GNU/Linux or Univention distribution in file
 # /usr/share/common-licenses/AGPL-3; if not, see
-# <http://www.gnu.org/licenses/>.
+# <https://www.gnu.org/licenses/>.
 
 import re
 from functools import wraps
+import random
 
+import six
 import ldap
 import ldap.schema
 import ldap.sasl
@@ -57,7 +59,6 @@ def parentDn(dn, base=''):
 	:return: The parent distinguished name or None.
 	:rtype: str or None
 	"""
-	_d = univention.debug.function('uldap.parentDn dn=%s base=%s' % (dn, base))  # noqa F841
 	if dn.lower() == base.lower():
 		return None
 	dn = ldap.dn.str2dn(dn)
@@ -105,7 +106,7 @@ def getRootDnConnection(start_tls=2, decode_ignorelist=[], reconnect=True):
 def getAdminConnection(start_tls=2, decode_ignorelist=[], reconnect=True):
 	# type: (int, List[str], bool) -> access
 	"""
-	Open a LDAP connection to the Master LDAP server using the admin credentials.
+	Open a LDAP connection to the Primary Directory Node LDAP server using the admin credentials.
 
 	:param int start_tls: Negotiate TLS with server. If `2` is given, the command will require the operation to be successful.
 	:param decode_ignorelist: List of LDAP attribute names which shall be handled as binary attributes.
@@ -124,7 +125,7 @@ def getAdminConnection(start_tls=2, decode_ignorelist=[], reconnect=True):
 def getBackupConnection(start_tls=2, decode_ignorelist=[], reconnect=True):
 	# type: (int, List[str], bool) -> access
 	"""
-	Open a LDAP connection to a Backup LDAP server using the admin credentials.
+	Open a LDAP connection to a Backup Directory Node LDAP server using the admin credentials.
 
 	:param int start_tls: Negotiate TLS with server. If `2` is given, the command will require the operation to be successful.
 	:param decode_ignorelist: List of LDAP attribute names which shall be handled as binary attributes.
@@ -146,8 +147,8 @@ def getBackupConnection(start_tls=2, decode_ignorelist=[], reconnect=True):
 		return access(host=backup, port=port, base=ucr['ldap/base'], binddn='cn=backup,' + ucr['ldap/base'], bindpw=bindpw, start_tls=start_tls, decode_ignorelist=decode_ignorelist, reconnect=reconnect)
 
 
-def getMachineConnection(start_tls=2, decode_ignorelist=[], ldap_master=True, secret_file="/etc/machine.secret", reconnect=True):
-	# type: (int, List[str], bool, str, bool) -> access
+def getMachineConnection(start_tls=2, decode_ignorelist=[], ldap_master=True, secret_file="/etc/machine.secret", reconnect=True, random_server=False):
+	# type: (int, List[str], bool, str, bool, bool) -> access
 	"""
 	Open a LDAP connection using the machine credentials.
 
@@ -156,7 +157,8 @@ def getMachineConnection(start_tls=2, decode_ignorelist=[], ldap_master=True, se
 	:type decode_ignorelist: list[str]
 	:param bool ldap_master: Open a connection to the Master if True, to the preferred LDAP server otherwise.
 	:param str secret_file: The name of a file containing the password credentials.
-	:param bool reconnect: Automatically reconect if the connection fails.
+	:param bool reconnect: Automatically reconnect if the connection fails.
+	:param bool random_server: Choose a random LDAP server from ldap/server/name and ldap/server/addition.
 	:return: A LDAP access object.
 	:rtype: univention.uldap.access
 	"""
@@ -166,25 +168,32 @@ def getMachineConnection(start_tls=2, decode_ignorelist=[], ldap_master=True, se
 	bindpw = open(secret_file).read().rstrip('\n')
 
 	if ldap_master:
-		# Connect to DC Master
+		# Connect to Primary Directory Node
 		port = int(ucr.get('ldap/master/port', '7389'))
 		return access(host=ucr['ldap/master'], port=port, base=ucr['ldap/base'], binddn=ucr['ldap/hostdn'], bindpw=bindpw, start_tls=start_tls, decode_ignorelist=decode_ignorelist, reconnect=reconnect)
 	else:
 		# Connect to ldap/server/name
 		port = int(ucr.get('ldap/server/port', '7389'))
-		try:
-			return access(host=ucr['ldap/server/name'], port=port, base=ucr['ldap/base'], binddn=ucr['ldap/hostdn'], bindpw=bindpw, start_tls=start_tls, decode_ignorelist=decode_ignorelist, reconnect=reconnect)
-		except ldap.SERVER_DOWN as exc:
-			# ldap/server/name is down, try next server
-			if not ucr.get('ldap/server/addition'):
-				raise
-			servers = ucr.get('ldap/server/addition', '')
-			for server in servers.split():
-				try:
-					return access(host=server, port=port, base=ucr['ldap/base'], binddn=ucr['ldap/hostdn'], bindpw=bindpw, start_tls=start_tls, decode_ignorelist=decode_ignorelist, reconnect=reconnect)
-				except ldap.SERVER_DOWN:
-					pass
-			raise exc
+		servers = [ucr.get('ldap/server/name')]
+		additional_servers = ucr.get('ldap/server/addition', '').split()
+		if random_server:
+			if ucr.get('server/role') in ('memberserver',) and random_server:
+				# shuffle all servers on Managed Nodes if random_server==True
+				servers += additional_servers
+				random.shuffle(servers)
+			else:
+				# shuffle only additional server
+				random.shuffle(additional_servers)
+				servers += additional_servers
+		else:
+			servers += additional_servers
+		for server in servers:
+			try:
+				return access(host=server, port=port, base=ucr['ldap/base'], binddn=ucr['ldap/hostdn'], bindpw=bindpw, start_tls=start_tls, decode_ignorelist=decode_ignorelist, reconnect=reconnect)
+			# LDAP server down, try next server
+			except ldap.SERVER_DOWN as oexc:
+				exc = oexc
+		raise exc
 
 
 def _fix_reconnect_handling(func):
@@ -232,7 +241,11 @@ class access(object):
 		self.host = host
 		self.base = base
 		self.binddn = binddn
+		if six.PY2 and isinstance(self.binddn, bytes):
+			self.binddn = self.binddn.decode('UTF-8')
 		self.bindpw = bindpw
+		if six.PY2 and isinstance(self.bindpw, bytes):
+			self.bindpw = self.bindpw.decode('UTF-8')
 		self.start_tls = start_tls
 		self.ca_certfile = ca_certfile
 		self.reconnect = reconnect
@@ -274,12 +287,6 @@ class access(object):
 
 		self.__open(ca_certfile)
 
-	def __encode_pwd(self, pwd):
-		if isinstance(pwd, unicode):
-			return str(pwd)
-		else:
-			return pwd
-
 	@_fix_reconnect_handling
 	def bind(self, binddn, bindpw):
 		# type: (str, str) -> None
@@ -292,7 +299,7 @@ class access(object):
 		self.binddn = binddn
 		self.bindpw = bindpw
 		univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'bind binddn=%s' % self.binddn)
-		self.lo.simple_bind_s(self.binddn, self.__encode_pwd(self.bindpw))
+		self.lo.simple_bind_s(self.binddn, self.bindpw)
 
 	@_fix_reconnect_handling
 	def bind_saml(self, bindpw):
@@ -328,7 +335,7 @@ class access(object):
 		:rtype: str
 		"""
 		dn = self.lo.whoami_s()
-		return re.sub('^dn:', '', dn)
+		return re.sub(u'^dn:', u'', dn)
 
 	def _reconnect(self):
 		# type: () -> None
@@ -337,7 +344,6 @@ class access(object):
 
 	def __open(self, ca_certfile):
 		# type: (Optional[str]) -> None
-		_d = univention.debug.function('uldap.__open host=%s port=%d base=%s' % (self.host, self.port, self.base))  # noqa F841
 
 		if self.reconnect:
 			univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'establishing new connection with retry_max=%d' % self. client_connection_attempt)
@@ -353,7 +359,7 @@ class access(object):
 			if self.start_tls == 1:
 				try:
 					self.__starttls()
-				except:
+				except Exception:
 					univention.debug.debug(univention.debug.LDAP, univention.debug.WARN, 'Could not start TLS')
 			elif self.start_tls == 2:
 				self.__starttls()
@@ -373,9 +379,11 @@ class access(object):
 		self.lo.start_tls_s()
 
 	def __encode(self, value):
+		if six.PY3:
+			return value
 		if value is None:
 			return value
-		elif isinstance(value, unicode):
+		elif six.PY2 and isinstance(value, unicode):  # noqa: F821
 			return str(value)
 		elif isinstance(value, (list, tuple)):
 			return map(self.__encode, value)
@@ -383,11 +391,15 @@ class access(object):
 			return value
 
 	def __recode_attribute(self, attr, val):
+		if six.PY3:
+			return val
 		if attr in self.decode_ignorelist:
 			return val
 		return self.__encode(val)
 
 	def __recode_entry(self, entry):
+		if six.PY3:
+			return entry
 		if isinstance(entry, tuple) and len(entry) == 3:
 			return (entry[0], entry[1], self.__recode_attribute(entry[1], entry[2]))
 		elif isinstance(entry, tuple) and len(entry) == 2:
@@ -428,10 +440,9 @@ class access(object):
 		if dn:
 			try:
 				result = self.lo.search_s(dn, ldap.SCOPE_BASE, '(objectClass=*)', attr)
-			except ldap.NO_SUCH_OBJECT:
-				result = []
-			if result:
 				return self.__decode_entry(result[0][1])
+			except (ldap.NO_SUCH_OBJECT, LookupError):
+				pass
 		if required:
 			raise ldap.NO_SUCH_OBJECT({'desc': 'no object'})
 		return {}
@@ -453,14 +464,12 @@ class access(object):
 
 		.. warning:: when `required=True` it raises `ldap.NO_SUCH_OBJECT` even if the object exists but the attribute is not set
 		"""
-		_d = univention.debug.function('uldap.getAttr %s %s' % (dn, attr))  # noqa F841
 		if dn:
 			try:
 				result = self.lo.search_s(dn, ldap.SCOPE_BASE, '(objectClass=*)', [attr])
-			except ldap.NO_SUCH_OBJECT:
-				result = []
-			if result and attr in result[0][1]:
 				return result[0][1][attr]
+			except (ldap.NO_SUCH_OBJECT, LookupError):
+				pass
 		if required:
 			raise ldap.NO_SUCH_OBJECT({'desc': 'no object'})
 		return []
@@ -532,7 +541,6 @@ class access(object):
 		:raises ldap.NO_SUCH_OBJECT: Indicates the target object cannot be found.
 		:raises ldap.INAPPROPRIATE_MATCHING: Indicates that the matching rule specified in the search filter does not match a rule defined for the attribute's syntax.
 		"""
-		_d = univention.debug.function('uldap.searchDn filter=%s base=%s scope=%s unique=%d required=%d' % (filter, base, scope, unique, required))  # noqa F841
 		return [x[0] for x in self.search(filter, base, scope, ['dn'], unique, required, timeout, sizelimit, serverctrls, response)]
 
 	@_fix_reconnect_handling
@@ -552,45 +560,44 @@ class access(object):
 			attrs = {}
 		if policies is None:
 			policies = []
-		_d = univention.debug.function('uldap.getPolicies dn=%s policies=%s attrs=%s' % (  # noqa F841
-			dn, policies, attrs))
 		if not dn and not policies:  # if policies is set apply a fictionally referenced list of policies
 			return {}
 
 		# get current dn
-		if attrs and 'objectClass' in attrs and 'univentionPolicyReference' in attrs:
+		if 'objectClass' in attrs and 'univentionPolicyReference' in attrs:
 			oattrs = attrs
 		else:
 			oattrs = self.get(dn, ['univentionPolicyReference', 'objectClass'])
-		if attrs and 'univentionPolicyReference' in attrs:
-			policies = attrs['univentionPolicyReference']
+
+		if 'univentionPolicyReference' in attrs:
+			policies = [x.decode('utf-8') for x in attrs['univentionPolicyReference']]
 		elif not policies and not attrs:
-			policies = oattrs.get('univentionPolicyReference', [])
+			policies = [x.decode('utf-8') for x in oattrs.get('univentionPolicyReference', [])]
 
 		object_classes = set(oc.lower() for oc in oattrs.get('objectClass', []))
 
-		result = {}
+		merged = {}
 		if dn:
 			obj_dn = dn
 			while True:
-				for policy_dn in policies:
-					self._merge_policy(policy_dn, obj_dn, object_classes, result)
-				dn = self.parentDn(dn)
+				for policy_dn in policies or []:
+					self._merge_policy(policy_dn, obj_dn, object_classes, merged)
+				dn = self.parentDn(dn) or ''
 				if not dn:
 					break
 				try:
 					parent = self.get(dn, attr=['univentionPolicyReference'], required=True)
 				except ldap.NO_SUCH_OBJECT:
 					break
-				policies = parent.get('univentionPolicyReference', [])
+				policies = [x.decode('utf-8') for x in parent.get('univentionPolicyReference', [])]
 
 		univention.debug.debug(
 			univention.debug.LDAP, univention.debug.INFO,
-			"getPolicies: result: %s" % result)
-		return result
+			"getPolicies: result: %s" % merged)
+		return merged
 
 	def _merge_policy(self, policy_dn, obj_dn, object_classes, result):
-		# type: (str, str, Set[str], Dict[str, Dict[str, Any]]) -> None
+		# type: (str, str, Set[bytes], Dict[str, Dict[str, Any]]) -> None
 		"""
 		Merge policies into result.
 
@@ -604,14 +611,14 @@ class access(object):
 			return
 
 		try:
-			classes = set(pattrs['objectClass']) - set(('top', 'univentionPolicy', 'univentionObject'))
-			ptype = classes.pop()
+			classes = set(pattrs['objectClass']) - {b'top', b'univentionPolicy', b'univentionObject'}
+			ptype = classes.pop().decode('utf-8')
 		except KeyError:
 			return
 
 		if pattrs.get('ldapFilter'):
 			try:
-				self.search(pattrs['ldapFilter'][0], base=obj_dn, scope='base', unique=True, required=True)
+				self.search(pattrs['ldapFilter'][0].decode('utf-8'), base=obj_dn, scope='base', unique=True, required=True)
 			except ldap.NO_SUCH_OBJECT:
 				return
 
@@ -620,22 +627,20 @@ class access(object):
 		if any(oc.lower() in object_classes for oc in pattrs.get('prohibitedObjectClasses', [])):
 			return
 
-		fixed = set(pattrs.get('fixedAttributes', ()))
-		empty = set(pattrs.get('emptyAttributes', ()))
+		fixed = set(x.decode('utf-8') for x in pattrs.get('fixedAttributes', ()))
+		empty = set(x.decode('utf-8') for x in pattrs.get('emptyAttributes', ()))
 		values = result.setdefault(ptype, {})
-		for key in list(empty) + pattrs.keys() + list(fixed):
-			if key in ('requiredObjectClasses', 'prohibitedObjectClasses', 'fixedAttributes', 'emptyAttributes', 'objectClass', 'cn', 'univentionObjectType', 'ldapFilter'):
-				continue
-
+		SKIP = {'requiredObjectClasses', 'prohibitedObjectClasses', 'fixedAttributes', 'emptyAttributes', 'objectClass', 'cn', 'univentionObjectType', 'ldapFilter'}
+		for key in (empty | set(pattrs) | fixed) - SKIP:
 			if key not in values or key in fixed:
 				value = [] if key in empty else pattrs.get(key, [])
 				univention.debug.debug(
 					univention.debug.LDAP, univention.debug.INFO,
-					"getPolicies: %s sets: %s=%s" % (policy_dn, key, value))
+					"getPolicies: %s sets: %s=%r" % (policy_dn, key, value))
 				values[key] = {
 					'policy': policy_dn,
 					'value': value,
-					'fixed': 1 if key in fixed else 0,
+					'fixed': key in fixed,
 				}
 
 	@_fix_reconnect_handling
@@ -676,10 +681,10 @@ class access(object):
 			key, val = i[0], i[-1]
 			if not val:
 				continue
-			if isinstance(val, basestring):
+			if isinstance(val, (bytes, six.text_type)):
 				val = [val]
-			nal.setdefault(key, set())
-			nal[key] |= set(val)
+			vals = nal.setdefault(key, set())
+			vals |= set(val)
 
 		nal = self.__encode_entry([(k, list(v)) for k, v in nal.items()])
 
@@ -716,12 +721,10 @@ class access(object):
 		ml = []
 		for key, oldvalue, newvalue in changes:
 			if oldvalue and newvalue:
-				if oldvalue == newvalue or (not isinstance(oldvalue, basestring) and not isinstance(newvalue, basestring) and set(oldvalue) == set(newvalue)):
+				if oldvalue == newvalue or (not isinstance(oldvalue, (bytes, six.text_type)) and not isinstance(newvalue, (bytes, six.text_type)) and set(oldvalue) == set(newvalue)):
 					continue  # equal values
 				op = ldap.MOD_REPLACE
 				val = newvalue
-				if (key == 'krb5ValidEnd' or key == 'krb5PasswordEnd') and newvalue == '0':  # TODO: move into the specific handlers
-					val = 0
 			elif not oldvalue and newvalue:
 				op = ldap.MOD_ADD
 				val = newvalue
@@ -731,7 +734,7 @@ class access(object):
 				# These attributes don't have a matching rule:
 				#   https://forge.univention.org/bugzilla/show_bug.cgi?id=15171
 				#   https://forge.univention.org/bugzilla/show_bug.cgi?id=44019
-				if key in ['jpegPhoto', 'univentionPortalBackground', 'univentionPortalLogo', 'univentionPortalEntryIcon', 'univentionUMCIcon']:
+				if key in ['preferredDeliveryMethod', 'jpegPhoto', 'univentionPortalBackground', 'univentionPortalLogo', 'univentionPortalEntryIcon', 'univentionUMCIcon']:
 					val = None
 			else:
 				continue
@@ -755,15 +758,17 @@ class access(object):
 		>>> get_dn = access._access__get_new_dn
 		>>> get_dn('univentionAppID=foo,dc=bar', [(ldap.MOD_REPLACE, 'univentionAppID', 'foo')])[0]
 		'univentionAppID=foo,dc=bar'
-		>>> get_dn('univentionAppID=foo,dc=bar', [(ldap.MOD_REPLACE, 'univentionAppID', 'föo')])[0]
-		'univentionAppID=f\\xc3\\xb6o,dc=bar'
+		>>> get_dn('univentionAppID=foo,dc=bar', [(ldap.MOD_REPLACE, 'univentionAppID', 'föo')])[0] == u'univentionAppID=föo,dc=bar'
+		True
 		>>> get_dn('univentionAppID=foo,dc=bar', [(ldap.MOD_REPLACE, 'univentionAppID', 'bar')])[0]
 		'univentionAppID=bar,dc=bar'
 		"""
 		rdn = ldap.dn.str2dn(dn)[0]
 		dn_vals = dict((x[0].lower(), x[1]) for x in rdn)
-		new_vals = dict((key.lower(), val if isinstance(val, basestring) else val[0]) for op, key, val in ml if val and op not in (ldap.MOD_DELETE,))
-		new_rdn = ldap.dn.dn2str([[(x, new_vals.get(x.lower(), dn_vals[x.lower()]), ldap.AVA_STRING) for x in [y[0] for y in rdn]]])
+		new_vals = dict((key.lower(), val if isinstance(val, (bytes, six.text_type)) else val[0]) for op, key, val in ml if val and op not in (ldap.MOD_DELETE,))
+		new_rdn_ava = [(x, new_vals.get(x.lower(), dn_vals[x.lower()]), ldap.AVA_STRING) for x in [y[0] for y in rdn]]
+		new_rdn_unicode = [(key, val.decode('UTF-8'), ava_type) if isinstance(val, bytes) else (key, val, ava_type) for key, val, ava_type in new_rdn_ava]
+		new_rdn = ldap.dn.dn2str([new_rdn_unicode])
 		rdn = ldap.dn.dn2str([rdn])
 		if rdn != new_rdn:
 			return ldap.dn.dn2str([ldap.dn.str2dn(new_rdn)[0]] + ldap.dn.str2dn(dn)[1:]), new_rdn
@@ -831,12 +836,12 @@ class access(object):
 		if not serverctrls:
 			serverctrls = []
 
-		if not newsdn.lower() == oldsdn.lower():
-			univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'uldap.rename: move %s to %s in %s' % (dn, newrdn, newsdn))
-			self.rename_ext_s(dn, newrdn, newsdn, serverctrls=serverctrls, response=response)
-		else:
+		if oldsdn and newsdn.lower() == oldsdn.lower():
 			univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'uldap.rename: modrdn %s to %s' % (dn, newrdn))
 			self.rename_ext_s(dn, newrdn, serverctrls=serverctrls, response=response)
+		else:
+			univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'uldap.rename: move %s to %s in %s' % (dn, newrdn, newsdn))
+			self.rename_ext_s(dn, newrdn, newsdn, serverctrls=serverctrls, response=response)
 
 	@_fix_reconnect_handling
 	def rename_ext_s(self, dn, newrdn, newsuperior=None, serverctrls=None, response=None):
@@ -945,7 +950,6 @@ class access(object):
 		"""
 		Return state for pickling.
 		"""
-		_d = univention.debug.function('uldap.__getstate__')  # noqa F841
 		odict = self.__dict__.copy()
 		del odict['lo']
 		return odict
@@ -954,7 +958,6 @@ class access(object):
 		"""
 		Set state for pickling.
 		"""
-		_d = univention.debug.function('uldap.__setstate__')  # noqa F841
 		self.__dict__.update(dict)
 		self.__open(self.ca_certfile)
 
@@ -983,12 +986,12 @@ class access(object):
 			if self.start_tls == 1:
 				try:
 					lo_ref.start_tls_s()
-				except:
+				except Exception:
 					univention.debug.debug(univention.debug.LDAP, univention.debug.WARN, 'Could not start TLS')
 			elif self.start_tls == 2:
 				lo_ref.start_tls_s()
 
-			lo_ref.simple_bind_s(self.binddn, self.__encode_pwd(self.bindpw))
+			lo_ref.simple_bind_s(self.binddn, self.bindpw)
 			return lo_ref
 
 		else:
